# Copyright(c) 2011-2012 Association of Universities for Research in Astronomy, Inc.

procedure gsat (inimage, extension)

# Return the saturation for the requested extension, adjusted for overscan
# where required.

# This task only accepts one input file at a time

# Saturation values in ADU, for EEV CCDs are stored internally to this
# script. Though currently there is only one value for all amplifiers.
# Full well depths in electrons for e2vDD CCDs are stored internally to this
# script. The values do not include bias values. One value per amplifier.
# This should be the case for Hamamatsu CCDs.

char    inimage    {prompt="Input image"}
char    extension  {prompt="Extension to return the saturation value for"}
char    gaindb     {"default", prompt="Database with gain data"}
char    bias       {"static", prompt="Bias value to use if image not overscan subtracted (static|calc)"}
char    pixstat    {"midpt", prompt="Statistic to be calculated by imstatistics to use to determine the rough BIAS level"}
char    statsec    {"default", prompt="Relative section of BIASSEC to estimate bias if bias==\"calc\". ([%d:%d,%d:%d])"}
char    gainval    {"default", prompt="Gain value for requested extension"}
real    saturation {0., prompt="Returned saturation value for requested extension"}
char    logfile    {"", prompt="Logfile for task"}
bool    verbose    {no, prompt="Verbose: yes or no"}
int     status     {0, prompt="Output status of task. 0 = GOOD, 1 = ERROR"}

begin

    char    filename, readmode_test_val, l_readmode, l_gainmode, l_key_gain
    char    filephu
    string  gaindbname, imstatsec, l_ampname
    real    gainmode_test_val, l_valgain, l_valbias, l_dbgain, flat_saturation
    real    l_gorig, controller_limit, ncomb
    int     bx1, bx2, by1, by2, ubx1, ubx2, uby1, uby2, nbiascontam
    int     xbin, ybin, counter
    bool    iseev, ise2vDD, ishamamatsu, inelectrons, isoversub, gainorig
    char    e2v_ampnames[6]="e2v 10031-23-05, left","e2v 10031-23-05, right","e2v 10031-01-03, left","e2v 10031-01-03, right","e2v 10031-18-04, left","e2v 10031-18-04, right"
    real    e2v_FWS[6]=110900.0,105100.0,115500.0,108900.0,115700.0,109200.0
    # full--well saturation for e2vDD CCDs 1x1 binning in electrons

    # Declare local user variables
    char    l_inimage
    char    l_extension
    char    l_gaindb
    char    l_bias
    char    l_pixstat
    char    l_statsec
    char    l_ugain
    real    l_saturation
    char    l_logfile
    bool    l_verbose

    # Read in user parameters
    l_inimage = inimage
    l_extension = extension
    l_gaindb = gaindb
    l_bias = bias
    l_pixstat = pixstat
    l_statsec = statsec
    l_ugain = gainval
    l_saturation = 0.0
    l_logfile = logfile
    l_verbose = verbose

    # Set the logfile for task
    if (l_logfile == "") {
        l_logfile = gmos.logfile
        if (l_logfile == "") {
            l_logfile = "gmos.log"
            printlog ("WARNING - GSAT: both gireduce.logfile and "//\
                "gmos.logfile are empty.", l_logfile, l_verbose)
            printlog ("                Using default file gmos.log.", \
                l_logfile, l_verbose)
        }
    }

    # Print start time
    gemdate (zone="local")
    printlog ("GSAT - Started: "//gemdate.outdate, l_logfile, l_verbose)

    #### Set default values ####

    # Booleans for dettector type
    iseev = no 
    ise2vDD = no
    ishamamatsu = no

    # Booleans for units / state of image / giflat convertions
    inelectrons = no
    isoversub = no
    gainorig = no

    # Used later on
    readmode_test_val = "1000" # Fast readout
    gainmode_test_val = 3.0 # Gain differentiator >3 high, <3 low.

    # Set the default controller digitization limit in ADU (including bias)
    controller_limit = 65535.0 # Zero-based

    # Reset saturation to 0. as it is used to return the value
    saturation = 0.

    # Reset status to 0, assume it's going to be OK until it's not
    status = 0

    #### Check input parameters and set other appropriately ####

    # Check that the input has ".fits" at the end
    if (substr(l_inimage,(strlen(l_inimage) - 4), strlen(l_inimage)) != \
        ".fits") {
        l_inimage = l_inimage//".fits"
    }

    # Check the input file exists
    if (!access(l_inimage)) {
        printlog ("ERROR - GSAT: Cannot access file "//l_inimage, \
            l_logfile, l_verbose)
        status = 1
        goto crash
    }

    # Set the phu variable
    filephu = l_inimage//"[0]"

    # Check the extension requested
    if ((l_extension == "") || (stridx (l_extension, " ") > 0)) {
        printlog ("ERROR - GSAT: extension has not been set", \
            l_logfile, l_verbose)
        status = 1
        goto crash
    } else {
        # Check whether the first and last characters of extension are "[" and
        # "]", respectively.
        if (substr(l_extension,1,1) != "[") {
            l_extension = "["//l_extension
        }

        if (substr(l_extension,strlen(l_extension),strlen(l_extension)) != \
            "]") {
            l_extension = l_extension//"]"
        }
        filename = l_inimage//l_extension
        # Check you can access the extension
        if (!imaccess(filename)) {
            printlog ("ERROR - GSAT: Cannot acces "//filename, \
                l_logfile, l_verbose)
            status = 1
            goto crash
        }
    }

    # Check gain database
    if ((l_gaindb == "") || (stridx (l_gaindb, " ") > 0)) {
        printlog ("ERROR - GSAT: gaindb has not been set", \
            l_logfile, l_verbose)
        status = 1
        goto crash
    } else if (l_gaindb == "default") {
        # Standard gain database - use ggdbhelper
        ggdbhelper (filephu, logfile=l_logfile)
        if (ggdbhelper.status != 0) {
            printlog ("ERROR - GSAT: GGDBHELPER returned a non-zero status", \
                l_logfile, l_verbose)
            status = 1
            goto crash
        }
        gaindbname = osfn(ggdbhelper.gaindb)
        if (!access(gaindbname)) {
            printlog ("ERROR - GSAT: Gain database "//gaindbname//\
                " not found", l_logfile, l_logfile)
            goto crash
        }
    } else {
        # User set gain database
        # Check it exists
        gaindbname = osfn(l_gaindb)
        if (!(access(gaindbname))) {
            printlog ("ERROR - GSAT: gaindb "//l_gaindb//\
                " cannot be accessed", l_logfile, l_verbose)
            status = 1
            goto crash
        }
    }

    # Check the bias parameter
    if (l_bias != "static" && l_bias != "calc") {
        printlog ("ERROR - GSAT: bias parameter must be set to \"static\""//\
            " or \"calc\"", l_logfile, l_verbose)
        status = 1
        goto crash
    }

    # Check the pixstat parameter
    if ((l_pixstat == "") || (stridx (l_pixstat, " ") > 0)) {
        printlog ("ERROR - GSAT: pixstat parameter not set", \
            l_logfile, l_verbose)
        status = 1
        goto crash
    }

    # Check the statsec parameter is set
    if ((l_statsec == "") || (stridx (l_statsec, " ") > 0)) {
        printlog ("ERROR - GSAT: gaindb has not been set", \
            l_logfile, l_verbose)
        status = 1
        goto crash
    }

    #### Perform the calculation of the saturation ####

    #### Read / set required information

    # Read DETTYPE
    keypar (filephu, "DETTYPE", silent+)
    if (!keypar.found) {
        printlog ("ERROR - GSAT: DETTYPE keywpord not found", \
            l_logfile, l_verbose)
        status = 1
        goto crash
    }

    if (keypar.value == "SDSU II CCD") {
        # EEV CCDs
        iseev = yes
        nbiascontam = 4
    } else if (keypar.value == "SDSU II e2v DD CCD42-90") {
        # e2vDD CCDs
        ise2vDD = yes
        nbiascontam = 5
    } else if (keypar.value == "S10892-01") { 
        #Hamamatsu CCDs
        ishamamatsu = yes
        nbiascontam = 4
        controller_limit = 65535 ##M
        readmode_test_val = "2000" ##M # Fast readout
        gainmode_test_val = 3.0 ##M
    }

    # Check if image is gtiled or gmosaiced - exit if e2vDD for now
    ##M For the future create a saturation mask on the fly / use DQ plain?
    keypar (filephu, "GMOSAIC", silent+)
    if (keypar.found && ise2vDD) {
        printlog ("ERROR - GSAT: GMOSAIC'ed images are not "//\
            "supported yet.", l_logfile, l_verbose)
        status = 1
        goto crash
    }
    keypar (filephu, "GTILE", silent+)
    if (keypar.found  && ise2vDD) {
        printlog ("ERROR - GSAT: GTILE'd images are not "//\
            "supported yet.", l_logfile, l_verbose)
        status = 1
        goto crash
    }

    # Obtain the AMPNAME
    keypar (filename, "AMPNAME", silent+)
    if (!keypar.found) {
        printlog ("ERROR - GSAT: Cannot find AMPNAME"//\
            "in "//filename, l_logfile, l_verbose)
        status = 1
        goto crash
    }
    l_ampname = str(keypar.value)

    # Read the gain and determine the units of the image
    ## Can I just read gain and if it's equal to 1 it's in electrons? - 
    ## giflat does something a little different
    l_key_gain = "GAINMULT" # Exists when image is converted to electrons
    keypar (filename, l_key_gain, silent+)
    if (!keypar.found) {
        l_key_gain = "GAIN"
        keypar (filename, l_key_gain, silent+)
        if (!keypar.found) {
            printlog ("ERROR - GSAT: Cannot find "//l_key_gain//\
                "in "//filename, l_logfile, l_verbose)
            status = 1
            goto crash
        } else {
            l_valgain = real(keypar.value)

            # Check for GAINORIG...
            # This is because of GIFLAT. GIFLAT changes GAIN to the number
            # of flats in the combined flat, and sets GAINORIG to the original
            # GAIN. This may change in the future - MS
            l_key_gain = "GAINORIG"
            keypar (filename, l_key_gain, silent+)
            if (keypar.found) {
                gainorig = yes
                l_gorig = real(keypar.value)
                printlog ("GSAT: GAINORIG keyword found, value is "//l_gorig, \
                    l_logfile, l_verbose)

                # Up to this point the number of combine flats is stored in
                # l_valgain again due to GIFLAT
                ncomb = l_valgain
                printlog ("GSAT: Number of combined flats is: "//ncomb, \
                    l_logfile, l_verbose)

                # Update the gain value accordingly             
                l_valgain = l_valgain * l_gorig
                flat_saturation = l_valgain
                # Units are in electrons - due to GIFLAT
                printlog ("GSAT: Input image is in electrons", \
                    l_logfile, l_verbose)
                inelectrons = yes
            }
        }
    } else {
        l_valgain = real(keypar.value)
        # Units are in electrons
        printlog ("GSAT: Input image is in electrons", \
            l_logfile, l_verbose)
        inelectrons = yes
    }

    # Set the gain mode
    if(l_valgain > gainmode_test_val) {
        l_gainmode = "high"
    } else if(l_valgain < gainmode_test_val) {
        l_gainmode = "low"
    }

    # Set the read mode
    l_readmode = "slow"
    l_gainmode = "low"
    keypar (filephu, "AMPINTEG", silent+)
    if (!keypar.found) {
        printlog ("WARNING - GSAT: Cannot find CCD readmode for image "//\
            inimage, l_logfile, verbose+)
        printlog ("                    Assuming slow read", \
            l_logfile, verbose+)
        status = 1
        goto crash
    } else if (keypar.value == readmode_test_val) {
            l_readmode = "fast"
    }

    # Initiate variables so they can be checked after setting later on
    l_valbias = -99999.0 # Bias value
    l_dbgain = -99999.0 # Gain read from gain database

    # Read the gain and 'static' bias from gaindb
    match (l_readmode, gaindbname, stop-) | \
        match (l_gainmode, "STDIN", stop-) | \
        match (l_ampname, "STDIN", stop-) | \
        fields (files="STDIN", fields="3,5", lines="1-", quit_if_miss=no, \
            print_file_n=no) | scan (l_dbgain, l_valbias)

    # Check variables have been changed
    if (l_dbgain == -99999.0 || l_valbias == -99999.0) {
        printlog ("ERROR - GSAT: Cannot read gain and bias values from "//\
            gaindbname, l_logfile, verbose+) 
        status = 1
        goto crash
    }

    # Check if gprepared - if not set l_valgain to l_dbgain
    # GPREPARE updates the GAIN keyword value to the database value
    keypar (filephu, "GPREPARE", silent+)
    if (!keypar.found) {
        l_valgain = l_dbgain
    }

    # Set user set gainb value if given
    if (l_ugain != "default") {
        l_valgain = real(l_ugain)
    }

    # Check if the image have been bias corrected already -
    # Require the value regardless to adjust detector limit saturation
    # May need to do checks for gmosaic / gtile / trimmed etc....
    #     Check for the following keywords:
    #         OVSERSCAN
    #         GOVERSUB
    keypar (filename, "OVERSCAN", silent+) # From GIREDUCE
    if (keypar.found) {
        l_valbias = real(keypar.value)
        isoversub = yes
    } else {
        keypar (filename, "GOVERSUB", silent+) # From goversub
        if (keypar.found) {
            l_valbias = real(keypar.value)
            isoversub = yes
        } else {
            # Check for BIASIM and DARKIM - use static value, set oversub to yes
            keypar (filephu, "BIASIM", silent+)
            if (keypar.found) {
                printlog ("WARNING - GSAT: No overscan value found but "//\
                    filename//" is BIAS corrected. Setting bias level "//\
                    "value to static value from "//gaindbname, \
                    l_logfile, l_verbose)
                isoversub = yes
            } else {
                keypar (filephu, "DARKIM", silent+)
                if (keypar.found) {
                    printlog ("WARNING - GSAT: No overscan value found but "//\
                        filename//" is DARK corrected. Setting bias level "//\
                        "value to static value from "//gaindbname, \
                        l_logfile, l_verbose)
                    isoversub = yes
                }
            }
        }
    }

    # Calculate bias from overscan if requested and if not already overscan
    # subtracted
    if (l_bias == "calc" && !isoversub) {

        # Check if trimmed
        keypar (filename, "TRIMMED", silent+)
        if (keypar.found) {
            printlog ("ERROR - GSAT: "//filename//" trimmed but not bias"//\
                " / overscan corrected.", l_logfile, l_verbose)
            status = 1
            goto crash
        }

        printlog ("GSAT: Calculating bias level", \
            l_logfile, l_verbose)

        # Read the BIASSEC
        keypar (filename, "BIASSEC", silent+)
        if (!keypar.found) {
            printlog ("ERROR - GSAT: Cannot find BIASSEC"//\
                "in "//filename, l_logfile, l_verbose)
            status = 1
            goto crash
        }

        # Parse the biassec
        print (keypar.value) | scanf ("[%d:%d,%d:%d]", bx1, bx2, by1, by2)

        if (l_statsec == "default") {
            # Biassec is on the left
            if (bx1 == 1) {
                bx2 -= nbiascontam
                bx1 += 1
            } else {
                # Biassec is on the right
                bx1 += nbiascontam
                bx2 -= 1
            }
        } else {
            # User section relative to BIASSEC
            print (l_statsec) | scanf ("[%d:%d,%d:%d]", \
                ubx1, ubx2, uby1, uby2)

            bx2 = bx1 + ubx2
            bx1 += ubx1
            by2 = by1 + uby2
            by1 += uby1
        }

        # Set the section to run imstat on
        imstatsec = "["//str(bx1)//":"//str(bx2)//","//str(by1)//":"//\
            str(by2)//"]"

        printlog ("GSAT: Imstatistics section used of input file: "//\
           imstatsec, l_logfile, l_verbose)

        # Initiate variable to check later on
        l_valbias = -999.99

        # Run imstatistics on the 'new' biassec.
        imstatistics (filename//imstatsec, fields=l_pixstat, \
            lower=INDEF, upper=INDEF, nclip=0, lsigma=3.0, usigma=3.0, \
            binwidth=0.1, format=no, cache=no) | scan (l_valbias)

        printlog ("GSAT: Calculated bias level for "//filename//" is: "//\
            l_valbias, l_logfile, l_verbose)

        # Check the output was sensible
        if (l_valbias == -999.99) {
             printlog ("ERROR - GSAT: imstatistics didn't return a \
                 value. Exiting", l_logfile, verbose+)
             status = 1
             goto crash
        }
    }

    # Read the binning
    keypar (filename, "CCDSUM", silent+)
    if (!keypar.found) {
        printlog ("ERROR - GSAT: Cannot read CCDSUM in "//filename, \
            l_logfile, l_verbose)
        status = 1
        got crash
    }
    print (keypar.value) | scanf("%d %d", xbin, ybin)

    printlog ("GSAT: Bias level is "//l_valbias//" Gain level is "//\
        l_valgain, l_logfile, l_verbose)
    printlog ("GSAT: Bias corrected - "//isoversub, l_logfile, l_verbose)
    printlog ("GSAT: Input in electrons - "//inelectrons, l_logfile, l_verbose)
    printlog ("GSAT: Xbin is "//xbin//" Ybin is "//ybin, l_logfile, l_verbose)
    
    # Calculate the saturation value, in the correct units and adjusted
    # appropriately for bias level
    if (ise2vDD) {
        counter = 0
        # Find the matching ampname and read the FWD (in electrons)
        for (counter=1; counter <= 6; counter += 1) {
            if (l_ampname == e2v_ampnames[counter]) {
                # In electrons (not including bias)
                l_saturation = e2v_FWS[counter]
                break
            }
        }

        printlog ("GSAT: Amplifier is "//l_ampname//" 1x1 Full Well Depth "//\
            "is: "//l_saturation, l_logfile, l_verbose)

        # Correct for binning
        l_saturation = l_saturation * xbin * ybin

        ## I think that if gainorig then the following 2 booleans should always
        ## be true, i.e., they will not ever get into those loops. - MS

        # Correct for being a combined flat if required
        if (gainorig) {
            l_saturation = l_saturation * ncomb
        }

        # Correct for gain
        if (!inelectrons) {
            l_saturation = l_saturation / l_valgain
        }

        # Correct if not overscan corrected
        if (!isoversub) {
            l_saturation += l_valbias
        }

    } else {
        # This is a bit of a hack for now for the EEV CCDs - MS
        l_saturation = controller_limit # In ADU
#        l_saturation = 65000.0 # In ADU

        # Correct for overscan subtraction
        if (isoversub) {
             l_saturation -= l_valbias
        }

        # Convert to electrons if needed.
        if (inelectrons) {
            l_saturation = l_saturation * l_valgain
        }
    }

    # Correct controller saturation values for overscan correction
    if (isoversub) {
        # In ADU
        controller_limit -= l_valbias
    }

    # Correct controller saturation values if input in electrons
    if (inelectrons) {
        controller_limit = controller_limit * l_valgain
    }

    printlog ("GSAT: Calculated CCD saturation limit is "//l_saturation, \
        l_logfile, l_verbose)
    printlog ("GSAT: Calculated controller saturation limit is "//\
        controller_limit, l_logfile, l_verbose)

    # Compare controller limit to CCD limit to set output saturation
    saturation = min (l_saturation, controller_limit)

    # This is a complete hack for the extremely near future to get the
    # regression tests to pass to get scripts up to an operational state - MS
    if (!ise2vDD) {
        if (gainorig) {
            saturation = flat_saturation 
        } else {
            saturation = real(65000)
        }
    }        

    printlog ("GSAT: Output saturation is "//saturation, \
        l_logfile, l_verbose)

    goto clean

crash:

   # No clean up required

clean:

    # Print finish time
    gemdate (zone="local")
    printlog ("GSAT - Finished: "//gemdate.outdate, l_logfile, l_verbose)

    # Print output status
    if (status == 0) {
        printlog ("GSAT - Exit status: GOOD", l_logfile, l_verbose)
    } else {
        printlog ("GSAT - Exit status: ERROR", l_logfile, l_verbose)
    }

end
